@@ -25,6 +25,7 @@
 
 #include <cups/cups.h>
 #include <cups/ppd.h>
+#include <cups/file.h>
 #include <signal.h>
 #include <sys/wait.h>
 #include <errno.h>
@@ -212,10 +213,14 @@ main(int  argc,				/* I - Number of command-line args */
 		fit;			/* Fit output to default page size? */
   ppd_file_t	*ppd;			/* PPD file */
   ppd_size_t	*size;			/* Current page size */
+  cups_file_t	*fp;			/* Post-processing input file */
   int		pdf_pid,		/* Process ID for pdftops */
 		pdf_argc,		/* Number of args for pdftops */
 		pstops_pid,		/* Process ID of pstops filter */
 		pstops_pipe[2],		/* Pipe to pstops filter */
+		need_post_proc = 0,     /* Post-processing needed? */
+		post_proc_pid = 0,	/* Process ID of post-processing */
+		post_proc_pipe[2],	/* Pipe to post-processing */
 		wait_children,		/* Number of child processes left */
 		wait_pid,		/* Process ID from wait() */
 		wait_status,		/* Status from child */
@@ -531,24 +536,56 @@ main(int  argc,				/* I - Number of command-line args */
   pdf_argv[pdf_argc] = NULL;
 
  /*
-  * Execute "pdftops/gs | pstops"...
+  * Do we need post-processing of the PostScript output to work around bugs
+  * of the printer's PostScript interpreter?
+  */
+
+#ifdef HAVE_PDFTOPS
+  need_post_proc = 0;
+#else
+  need_post_proc =
+    (ppd && ppd->manufacturer &&
+     (!strncasecmp(ppd->manufacturer, "Kyocera", 7) ||
+      !strncasecmp(ppd->manufacturer, "Brother", 7)) ? 1 : 0);
+#endif /* HAVE_PDFTOPS */
+
+ /*
+  * Execute "pdftops/gs | pstops [ | post-processing ]"...
   */
 
   if (pipe(pstops_pipe))
   {
-    perror("DEBUG: Unable to create pipe");
+    perror("DEBUG: Unable to create pipe for pstops");
 
     exit_status = 1;
     goto error;
   }
 
+  if (need_post_proc)
+  {
+    if (pipe(post_proc_pipe))
+    {
+      perror("DEBUG: Unable to create pipe for post-processing");
+
+      exit_status = 1;
+      goto error;
+    }
+  }
+
   if ((pdf_pid = fork()) == 0)
   {
    /*
     * Child comes here...
     */
 
-    dup2(pstops_pipe[1], 1);
+    if (need_post_proc)
+    {
+      dup2(post_proc_pipe[1], 1);
+      close(post_proc_pipe[0]);
+      close(post_proc_pipe[1]);
+    }
+    else
+      dup2(pstops_pipe[1], 1);
     close(pstops_pipe[0]);
     close(pstops_pipe[1]);
 
@@ -580,12 +617,145 @@ main(int  argc,				/* I - Number of command-line args */
 
   fprintf(stderr, "DEBUG: Started filter %s (PID %d)\n", pdf_argv[0], pdf_pid);
 
+  if (need_post_proc)
+  {
+    if ((post_proc_pid = fork()) == 0)
+    {
+     /*
+      * Child comes here...
+      */
+
+      dup2(post_proc_pipe[0], 0);
+      close(post_proc_pipe[0]);
+      close(post_proc_pipe[1]);
+      dup2(pstops_pipe[1], 1);
+      close(pstops_pipe[0]);
+      close(pstops_pipe[1]);
+
+      fp = cupsFileStdin();
+
+     /*
+      * Copy everything until after initial comments (Prolog section)
+      */
+      while ((bytes = cupsFileGetLine(fp, buffer, sizeof(buffer))) > 0 &&
+	     strncmp(buffer, "%%BeginProlog", 13) &&
+	     strncmp(buffer, "%%EndProlog", 11) &&
+	     strncmp(buffer, "%%BeginSetup", 12) &&
+	     strncmp(buffer, "%%Page:", 7))
+	printf("%s", buffer);
+
+      if (bytes > 0)
+      {
+       /*
+	* Insert PostScript interpreter bug fix code in the beginning of
+	* the Prolog section (before the first active PostScript code)
+	*/
+	if (strncmp(buffer, "%%BeginProlog", 13))
+	{
+	  /* No Prolog section, create one */
+	  fprintf(stderr, "DEBUG: Adding Prolog section for workaround PostScript code\n");
+	  puts("%%BeginProlog");
+	}
+	else
+	  printf("%s", buffer);
+	
+#ifndef HAVE_PDFTOPS
+
+	if (ppd && ppd->manufacturer)
+	{
+
+	 /*
+	  * Kyocera printers have a bug in their PostScript interpreter 
+	  * making them crashing on PostScript input data generated by
+	  * Ghostscript's "ps2write" output device.
+	  *
+	  * The problem can be simply worked around by preceding the PostScript
+	  * code with some extra bits.
+	  *
+	  * See https://bugs.launchpad.net/bugs/951627
+	  */
+
+	  if (!strncasecmp(ppd->manufacturer, "Kyocera", 7))
+	  {
+	    puts("% ===== Workaround insertion by pdftops CUPS filter =====");
+	    fprintf(stderr, "DEBUG: Inserted workaround PostScript code for Kyocera printers\n");
+	    puts("% Kyocera's PostScript interpreter crashes on early name binding,");
+	    puts("% so eliminate all \"bind\"s by redifining \"bind\" to no-op");
+	    puts("/bind {} bind def");
+	    puts("% =====");
+	  }
+
+	 /*
+	  * Brother printers have a bug in their PostScript interpreter 
+	  * making them printing one blank page if PostScript input data
+	  * generated by Ghostscript's "ps2write" output device is used.
+	  *
+	  * The problem can be simply worked around by preceding the PostScript
+	  * code with some extra bits.
+	  *
+	  * See https://bugs.launchpad.net/bugs/950713
+	  */
+
+	  else if (!strncasecmp(ppd->manufacturer, "Brother", 7))
+	  {
+	    fprintf(stderr, "DEBUG: Inserted workaround PostScript code for Brother printers\n");
+	    puts("% ===== Workaround insertion by pdftops CUPS filter =====");
+	    puts("% Brother's PostScript interpreter spits out the current page");
+	    puts("% and aborts the job on the \"currenthalftone\" operator, so redefine");
+	    puts("% it to null");
+	    puts("/currenthalftone {//null} bind def");
+	    puts("/orig.sethalftone systemdict /sethalftone get def");
+	    puts("/sethalftone {dup //null eq not {//orig.sethalftone}{pop} ifelse} bind def");
+	    puts("% =====");
+	  }
+	}
+	
+#endif /* !HAVE_PDFTOPS */
+
+	if (strncmp(buffer, "%%BeginProlog", 13))
+	{
+	  /* Close newly created Prolog section */
+	  if (strncmp(buffer, "%%EndProlog", 11))
+	    puts("%%EndProlog");
+	  printf("%s", buffer);
+	}
+
+       /*
+	* Copy the rest of the file
+	*/
+	while ((bytes = cupsFileRead(fp, buffer, sizeof(buffer))) > 0)
+	  fwrite(buffer, 1, bytes, stdout);
+
+      }
+
+      exit(0);
+    }
+    else if (post_proc_pid < 0)
+    {
+     /*
+      * Unable to fork!
+      */
+
+      perror("DEBUG: Unable to execute post-processing process");
+
+      exit_status = 1;
+      goto error;
+    }
+
+    fprintf(stderr, "DEBUG: Started post-processing (PID %d)\n", post_proc_pid);
+  }
+
   if ((pstops_pid = fork()) == 0)
   {
    /*
     * Child comes here...
     */
 
+    if (need_post_proc)
+    {
+      close(post_proc_pipe[0]);
+      close(post_proc_pipe[1]);
+    }
     dup2(pstops_pipe[0], 0);
     close(pstops_pipe[0]);
     close(pstops_pipe[1]);
@@ -611,12 +781,17 @@ main(int  argc,				/* I - Number of command-line args */
 
   close(pstops_pipe[0]);
   close(pstops_pipe[1]);
+  if (need_post_proc)
+  {
+    close(post_proc_pipe[0]);
+    close(post_proc_pipe[1]);
+  }
 
  /*
   * Wait for the child processes to exit...
   */
 
-  wait_children = 2;
+  wait_children = 2 + need_post_proc;
 
   while (wait_children > 0)
   {
@@ -629,6 +804,8 @@ main(int  argc,				/* I - Number of command-line args */
       if (job_canceled)
       {
 	kill(pdf_pid, SIGTERM);
+	if (need_post_proc)
+	  kill(post_proc_pid, SIGTERM);
 	kill(pstops_pid, SIGTERM);
 
 	job_canceled = 0;
@@ -650,37 +827,40 @@ main(int  argc,				/* I - Number of command-line args */
       {
 	exit_status = WEXITSTATUS(wait_status);
 
-        fprintf(stderr, "DEBUG: PID %d (%s) stopped with status %d!\n",
-	        wait_pid,
+	fprintf(stderr, "DEBUG: PID %d (%s) stopped with status %d!\n",
+		wait_pid,
 #ifdef HAVE_PDFTOPS
-                wait_pid == pdf_pid ? "pdftops" : "pstops",
+		wait_pid == pdf_pid ? "pdftops" :
 #else
-                wait_pid == pdf_pid ? "gs" : "pstops",
+		wait_pid == pdf_pid ? "gs" :
 #endif /* HAVE_PDFTOPS */
+		(wait_pid == pstops_pid ? "pstops" : "Post-processing"),
 		exit_status);
       }
       else if (WTERMSIG(wait_status) == SIGTERM)
       {
-        fprintf(stderr,
-	        "DEBUG: PID %d (%s) was terminated normally with signal %d!\n",
-	        wait_pid,
+	fprintf(stderr,
+		"DEBUG: PID %d (%s) was terminated normally with signal %d!\n",
+		wait_pid,
 #ifdef HAVE_PDFTOPS
-                wait_pid == pdf_pid ? "pdftops" : "pstops",
+		wait_pid == pdf_pid ? "pdftops" :
 #else
-                wait_pid == pdf_pid ? "gs" : "pstops",
+		wait_pid == pdf_pid ? "gs" :
 #endif /* HAVE_PDFTOPS */
+		(wait_pid == pstops_pid ? "pstops" : "Post-processing"),
 		exit_status);
       }
       else
       {
 	exit_status = WTERMSIG(wait_status);
 
-        fprintf(stderr, "DEBUG: PID %d (%s) crashed on signal %d!\n", wait_pid,
+	fprintf(stderr, "DEBUG: PID %d (%s) crashed on signal %d!\n", wait_pid,
 #ifdef HAVE_PDFTOPS
-                wait_pid == pdf_pid ? "pdftops" : "pstops",
+		wait_pid == pdf_pid ? "pdftops" :
 #else
-                wait_pid == pdf_pid ? "gs" : "pstops",
+		wait_pid == pdf_pid ? "gs" :
 #endif /* HAVE_PDFTOPS */
+		(wait_pid == pstops_pid ? "pstops" : "Post-processing"),
 		exit_status);
       }
     }
@@ -688,10 +868,11 @@ main(int  argc,				/* I - Number of command-line args */
     {
       fprintf(stderr, "DEBUG: PID %d (%s) exited with no errors.\n", wait_pid,
 #ifdef HAVE_PDFTOPS
-	      wait_pid == pdf_pid ? "pdftops" : "pstops");
+	      wait_pid == pdf_pid ? "pdftops" :
 #else
-	      wait_pid == pdf_pid ? "gs" : "pstops");
+	      wait_pid == pdf_pid ? "gs" :
 #endif /* HAVE_PDFTOPS */
+	      (wait_pid == pstops_pid ? "pstops" : "Post-processing"));
     }
   }
 
